# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Sparse Tensor Representation.

See the [Sparse Ops](https://tensorflow.org/api_guides/python/sparse_ops) guide.

"""

from __future__ import print_function

from tensorflow.python import sparse_cross as cross
from tensorflow.python import sparse_cross_hashed as cross_hashed
from tensorflow.python import sparse_expand_dims as expand_dims
from tensorflow.python import sparse_eye as eye

del print_function
